RPL 

****************************************************************
* Fast Circle V2.5
* Algorithme de cration de cercle (Mid-Point)
* Exercice pour formation  l'assembleur SASM
* Yann Collet, 10/08/08
* Thanks to : Cyrille de Brebisson, for providing ideas & sample source code
****************************************************************
* Entre : Pile : 4 : Grob [Optional]  <-- Si non fourni, dessine dans le GRAPH (PICT) courant
*                 3 : Xcenter (en pixel) <= 1000
*                 2 : Ycenter (en pixel) <= 1000
*                 1 : Radius  (en pixel) <= 1000
****************************************************************

::
  CK3&Dispatch

  # 111
  ::
    ROTDUP %0< IT :: % 1048576 %+ ; COERCE   ( x )
    ROTDUP %0< IT :: % 1048576 %+ ; COERCE   ( y )
    ROT COERCE 1000 #MIN                     ( 1000 >= Radius >= 0 )

    DEPTH THREE #> ITE 4PICK DUP             ( Vrifie Stack4 = Grob )
    DUPTYPEGROB? ?SKIP :: DROP CHECKPICT GBUFF ;  ( Sinon, fourni GBUFF )
    4UNROLL
    
**********************************************************
* Principe : x + y <= R + MaxErr
*
* Initialise : MaxErr = R (Car [(R+1)-R]/2 = R + 1/2)
*	       On translate : MaxErr = 2R; Err = R (comme a on reste toujours positif)
*              On commence avec : x=R, y=0,
* Au lieu de calculer x+y, on va juste prendre en compte les Delta
*              NextDeltaX = 2R-1
*              NextDeltaY = 1
*
* Puis on augmente rcursivement : y=y+1
*              Err = Err + NextDeltaY
*              NextDeltaY = NextDeltaY + 2
*      Si Err > MaxErr, faut aussi diminuer x=x-1
*              Err = Err - NextDeltaX
*              NextDeltaX = NextDeltaX - 2
*      Puis on dessine
*      Jusqu' y>x
*
* Pour chaque point calcul, on dessine 8 pts :
* (x,y) (x,-y) (-x,y) (-x,-y) (y,x) (y,-x) (-y,x) (-y,-x)
*
**********************************************************
* MAP
* R0 : MaxY (X)
* R1 : x  (X)    Err         (M)
* R2 : x0 (X)    NextDeltaX  (M)
* R3 : y  (X)    MaxErr      (M)
* R4 : y0 (X)    NextDeltaY  (M)
* D0 : Multiplication code
* D1 : Grob.Pointer
**********************************************************

CODE

*** Cration de R1 > x (=R) & Err (=R);
  A=0 W
  GOSBVL =POP#           A.A = R = A.X ( Note : C.A & D.A modifis , D1 utilis)
  C=A W                  C.X = R (<=1000) = Err ; C.M = 0
  CSL A
  CSL A
  CSL W                  C.M = Err = R
  C=A X                  C.X = x = R
  R1=C

*** Cration de R3 > y (=0) & MaxErr = (2R)
		C=C+C M                C.M = MaxErr = 2R
		C=0 X                  C.X = y = 0		
  R3=C

*** Cration R4 ( y0 + NextDeltaY(=1) )
  GOSBVL =POP2#          A.A = X0 ; C.A = Y0
  A=R3.F M               A.M = 2R
		A=A-1 M                A.M = 2R-1 = NextDeltaX
  R2=A

*** Cration R2 ( x0 + NextDeltaX(=2R-1) )
  C=0 M
  CBIT=1 12              C.M = 1 = NextDeltaY
  R4=C

	 GOSBVL =SAVPTR         Sauve B.A, D.A, D0 & D1 (Donc ne pas les modifier!) mais dtruit C

*** Cration Zone Multiplication
		A=DAT1 A               Grob
  R0=A.F A               Sauvegarde Grob, car D0 & D1 vont tre modifis
  LC(5) #00028           C=40d
		GOSBVL =CREATETEMP     CREATETEMP : Allocates C.A nibbles carry if not enough memory
*                        -> D0=bottom, D1=top of area;
*                        -> B.A = C.A = @D1 = offset to previous tempob = #nibbles+6

*** MaxY & PointerGrob
		A=R0.F A               Grob
  D1=A
  D1=D1+ 10              MaxY
  A=DAT1 A               A.A = MaxY = A.X ( car MaxY<=1000 )
  R0=A.F X               Sauvegarde MaxY & NibbleWidth
  D1=D1+ 5               D1 Pointe sur MaxX

*** Cration Code Multiplication
	 A=0 W                    On nettoie A car il y a des shift plus tard
	 A=DAT1 A                 MaxX
  A=A+CON A,7         [8]  Pour passer l'octet suivant
  A=A+A A             [6]
  ASR A               [7]
  A=A+A A             [6]  A.A = NibbleWidth
	
	 C=D0
	 D=C A               Sauvegarde D0
		GOSHORT START_MULT
MULT_LOOP
	 DAT0=C B
	 D0=D0+ 2
START_MULT
	 ?A=0 A
	 GOYES END_MULT
		?A=0 P
		GOYES MULT_CSL
		?ABIT=0 0
		GOYES MULT_CC
ADD_BC
	 LC(2) #1C           B=B+C A (invers)
		ABIT=0 0            Mise  jour de A
	 GOSHORT MULT_LOOP
MULT_CSL
	 LC(2) #2F           CSL A (invers)
	 ASR A               Mise  jour de A
	 GOSHORT MULT_LOOP
MULT_CC
	 LC(2) #6C           C=C+C A (invers)
  ASRB              	 Mise  jour de A
	 GOSHORT MULT_LOOP

END_MULT
	 LC(4) #2B18         PC=A (invers)
	 DAT0=C 4

  C=D A
		D0=C                Restore D0	


**********************************************************
* MAP
* R0 : MaxY (X)
* R1 : x  (X)    Err         (M)
* R2 : x0 (X)    NextDeltaX  (M)
* R3 : y  (X)    MaxErr      (M)
* R4 : y0 (X)    NextDeltaY  (M)
* D0 : Multiplication code
* D1 : Grob.Pointer
**********************************************************
* Init pour DRAW : C.X=R1.X=x
  C=R1.F X
BOUCLE
  GOSUB DRAW_EIGHT_POINTS
  GOSUB Y_PLUS_ONE       Note : C=R1=X/Err
  A=R3.F X               A.X = Y
  ?C>=A X
  GOYES BOUCLE           Note : C=R1
  GOTO EXIT              Si Y>X, on arrte

Y_PLUS_ONE
  C=R3
  C=C+1 X           C.X = Y+1; C.M = MaxErr
  R3=C.F X          Sauvegarde Y+1
  A=R4.F M          A.M = NextDeltaY
  B=C M             B.M = MaxErr
  C=R1              C.M = Err; C.X = x
  C=C+A M           C.M = NewErr
  A=A+CON M,2       A.M = New.DeltaY
  R4=A.F M          Sauvegarde NewDeltaY

TEST_X_MINUS_ONE
  ?C<=B M
  GOYES ENDTESTX
X_MINUS_ONE
  A=R2.F M          A.M = NextDeltaX
  C=C-A M           C.M = NewErr
  A=A-CON M,2       A.M = New.DeltaX
  R2=A.F M          Sauvegarde NextDeltaX
  C=C-1 X           C.X = x-1
ENDTESTX
  R1=C              Sauvegarde x & Err
  RTN               Status : C=R1=X/Err


**********************************************************
* MAP
* R0 : MaxY (X)
* R1 : x  (X)    Err         (M)
* R2 : x0 (X)    NextDeltaX  (M)
* R3 : y  (X)    MaxErr      (M)
* R4 : y0 (X)    NextDeltaY  (M)
* D0 : Multiplication code
* D1 : Grob.Pointer
**********************************************************

DRAW_EIGHT_POINTS
  ST=1 0
DRAW_FOUR_POINTS
*  C=R1.F X                 Useless : C=R1
  B=C X                     B=x
  C=R3.F X
  D=C X                     D=y
  GOSUB CALCULATE_PIXEL

  C=R1.F X
  C=-C X
  B=C X                     B=-x
  C=R3.F X
  D=C X                     D=y
  GOSUB CALCULATE_PIXEL

  C=R1.F X
  B=C X                     B=x
  C=R3.F X
  C=-C X
  D=C X                     D=-y
  GOSUB CALCULATE_PIXEL

  C=R1.F X
  C=-C X
  B=C X                     B=-x
  C=R3.F X
  C=-C X
  D=C X                     D=-y
  GOSUB CALCULATE_PIXEL

  C=R1.F X                  C.X = x
  A=R3.F X                  A.X = y
  ACEX X                    C.X = y ; A.X = x
  R1=C.F X                  NewX = y;
  R3=A.F X                  NewY = x;

  ?ST=0 0
  RTNYES
  ST=0 0
  GOTO DRAW_FOUR_POINTS

**********************************************************
* MAP
* R0 : MaxY (X)
* R1 : x  (X)    Err         (M)
* R2 : x0 (X)    NextDeltaX  (M)
* R3 : y  (X)    MaxErr      (M)
* R4 : y0 (X)    NextDeltaY  (M)
* D0 : Multiplication code
* D1 : Grob.Pointer
**********************************************************
CALCULATE_PIXEL
* Recenter x,y using x0 & y0; Need : B=x, D=y;
  A=R2.F X            [9]  A.X = x0
  A=A+B X             [6]  A.X = Xpixel
  C=R4.F X            [9]  C.X = y0
  C=C+D X             [6]  C.X = Ypixel

CHECK_PIXEL
* Vrifie d'abord qu'on est bien dans la partie autorise de l'cran
  D=C X               [6]  D.X = y
  C=R0.F X            [9]  C.X = MaxY
  ?D>=C X             (9-11)
  RTNYES              Sortie si Y >= MaxY

  C=DAT1 A            [18] C.A = MaxX = C.X car MaxX<=1000
  ?A>=C X             (9-11)
  RTNYES              Sortie si X >= MaxX
  

DRAW_PIXEL
* On dessine un pixel  l'cran
* D1 sur GBUFF; A.X=x; B.X=NibbleWidth; C.X=(Free); D.X=y
  B=0 A               [7]
  B=B+CON A,5         [10] Offset, car D0 est sur MaxX

		C=A X
		CDEX X                   C = y ; On sauve x dans D

  A=A+A X             [6]  Note A.X = x <=1000, donc pas de dbordement
  A=A+A X             [6]
		ASR X               [7]  A.X / 4
  B=B+A X             [7]  Nibble dans la ligne

		A=D0                [16] Adresse Code Multiplication
	 APCEX               [16] Va  la multiplication; nibble x C.A ajout  B.A

  A=D1                [16] A=D1 c'est en fait AD1EX(8);D1=A(8)
  A=A+B A             [7]
  AD1EX               [8]  D1 pointe sur le bon nibble; A = GBUFF

* Allumage du bon pixel
* Technique de slection du masque utilisant P, par Detlef Muller

  LC(1) #C            [3]  A.0 = #Bh = 1100b
  C=C!D P             [5]  C.0 = 12/13/14/15 (P=0)
  P=C 0               [6]
  LC(4) #1248         [6]  C.S = Mask
  P= 0                [2]

  A=DAT1  S           [18]
  A=A!C   S           [5]  Trigger pixel
  DAT1=A  S           [17]

  D1=A                [8]
  
  RTN                 [9]

EXIT
  GOVLNG =GETPTRLOOP
ENDCODE

		DROP
  ;
;


